﻿using Infragistics.Documents.Parsing;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Shapes;

namespace TSqlCasingWorkaround
{
	/// <summary>
	/// Interaction logic for MainWindow.xaml
	/// </summary>
	public partial class MainWindow : Window
	{
		public MainWindow()
		{
			InitializeComponent();

			var tsqlLanguage = TSqlLanguage.Instance;
			var customGrammar = tsqlLanguage.Grammar.Clone();

			// Replace all LiteralIgnoreCase terminal symbols with equivalent RegularExpression symbols.
			ReplaceLiteralIgnoreCaseSymbols(customGrammar.LexerStates.DefaultLexerState);
			foreach (var lexerState in customGrammar.LexerStates)
				ReplaceLiteralIgnoreCaseSymbols(lexerState);

			// Create a language based on our cloned grammar and copy over the services from the T-SQL language.
			var customLanguage = new CustomLanguage(customGrammar);
			foreach (var serviceKey in tsqlLanguage.ServicesManager.GetRegisteredServiceKeys())
				customLanguage.ServicesManager.RegisterService(serviceKey, tsqlLanguage.ServicesManager.GetService(serviceKey));

			foreach (var serviceType in tsqlLanguage.ServicesManager.GetRegisteredServiceTypes())
				customLanguage.ServicesManager.RegisterService(serviceType, tsqlLanguage.ServicesManager.GetService(serviceType));

			// Set the language on the syntax editor.
			this.syntaxEditor.Document.Language = customLanguage;
		}

		private static Dictionary<NonTerminalSymbol, List<Tuple<SyntaxRule, List<int>>>> GetNonTerminalsOwningSymbol(Symbol symbol)
		{
			var nonTerminals = new Dictionary<NonTerminalSymbol, List<Tuple<SyntaxRule, List<int>>>>();

			foreach (var nonTerminal in symbol.Grammar.NonTerminalSymbols)
			{
				if (nonTerminal.Rule.Type == SyntaxRuleType.SymbolReference)
				{
					if (((SymbolReferenceSyntaxRule)nonTerminal.Rule).Symbol == symbol)
						nonTerminals.Add(nonTerminal, null);
				}
				else
				{
					var rulesAndIndexes = new List<Tuple<SyntaxRule, List<int>>>();
					GetRulesOwningSymbolReference(symbol, nonTerminal.Rule, rulesAndIndexes);

					if (rulesAndIndexes.Count != 0)
						nonTerminals.Add(nonTerminal, rulesAndIndexes);
				}
			}

			return nonTerminals;
		}

		private static void GetRulesOwningSymbolReference(Symbol symbol, SyntaxRule parentRule, List<Tuple<SyntaxRule, List<int>>> rulesAndIndexes)
		{
			var ruleOwner = parentRule as SyntaxRuleOwnerBase;
			if (ruleOwner != null)
			{
				if (ruleOwner.Rule != null)
				{
					var indexes = new List<int>();

					if (ruleOwner.Rule.Type == SyntaxRuleType.SymbolReference)
					{
						if (((SymbolReferenceSyntaxRule)ruleOwner.Rule).Symbol == symbol)
							indexes.Add(0);
					}
					else
					{
						GetRulesOwningSymbolReference(symbol, ruleOwner.Rule, rulesAndIndexes);
					}

					if (ruleOwner.Type == SyntaxRuleType.Exception)
					{
						var exceptionRule = (ExceptionSyntaxRule)ruleOwner;
						if (exceptionRule.Exception.Type == SyntaxRuleType.SymbolReference)
						{
							if (((SymbolReferenceSyntaxRule)exceptionRule.Exception).Symbol == symbol)
								indexes.Add(1);
						}
						else
						{
							GetRulesOwningSymbolReference(symbol, exceptionRule.Exception, rulesAndIndexes);
						}
					}

					if (indexes.Count != 0)
						rulesAndIndexes.Add(Tuple.Create(parentRule, indexes));
				}

				return;
			}

			var rulesOwner = parentRule as SyntaxRulesOwnerBase;
			if (rulesOwner != null)
			{
				var indexes = new List<int>();
				for (int i = 0; i < rulesOwner.Rules.Count; i++)
				{
					var current = rulesOwner.Rules[i];
					if (current.Type == SyntaxRuleType.SymbolReference)
					{
						if (((SymbolReferenceSyntaxRule)current).Symbol == symbol)
							indexes.Add(i);
					}
					else
					{
						GetRulesOwningSymbolReference(symbol, current, rulesAndIndexes);
					}
				}

				if (indexes.Count != 0)
					rulesAndIndexes.Add(Tuple.Create(parentRule, indexes));

				return;
			}

			Debug.Fail("Unexpected");
		}

		private static void ReplaceLiteralIgnoreCaseSymbols(LexerState lexerState)
		{
			for (int i = 0; i < lexerState.Symbols.Count; i++)
			{
				var symbol = lexerState.Symbols[i];
				if (symbol.Comparison != TerminalSymbolComparison.LiteralIgnoreCase)
					continue;

				var lexerStateToEnter = symbol.LexerStateToEnter;
				var isExitSymbol = lexerState.Symbols.IsExitSymbol(symbol);

				var nonTerminalsOwningSymbol = GetNonTerminalsOwningSymbol(symbol);

				lexerState.Symbols.Remove(symbol);
				var replacedSymbol = lexerState.Symbols.Insert(i,
					symbol.Name,
					"(?i)" + Regex.Escape(symbol.Value),
					TerminalSymbolComparison.RegularExpression,
					isExitSymbol);

				replacedSymbol.IsLookaheadNegative = symbol.IsLookaheadNegative;
				replacedSymbol.IsSignificant = symbol.IsSignificant;
				replacedSymbol.IsStartOfErrorRecoveryPair = symbol.IsStartOfErrorRecoveryPair;
				replacedSymbol.LanguageElementName = symbol.LanguageElementName;
				replacedSymbol.LexerStateToEnter = lexerStateToEnter;

				if (string.IsNullOrEmpty(symbol.LookaheadPattern) == false)
					replacedSymbol.LookaheadPattern = symbol.LookaheadPattern;

				ReplaceSymbolInNonTerminals(replacedSymbol, nonTerminalsOwningSymbol);
			}
		}

		private static void ReplaceSymbolInNonTerminals(Symbol replacedSymbol, Dictionary<NonTerminalSymbol, List<Tuple<SyntaxRule, List<int>>>> oldReferences)
		{
			foreach (var pair in oldReferences)
			{
				if (pair.Value == null)
				{
					Debug.Assert(pair.Key.Rule == null, "The rule should have been removed.");
					pair.Key.Rule = replacedSymbol;
					continue;
				}

				foreach (var tuple in pair.Value)
				{
					var rule = tuple.Item1;
					var indexes = tuple.Item2;

					var ruleOwner = rule as SyntaxRuleOwnerBase;
					if (ruleOwner != null)
					{
						Debug.Assert(indexes.Count <= 2, "Unexpected number of indexes.");

						if (indexes.Contains(0))
						{
							Debug.Assert(ruleOwner.Rule == null, "The rule should have been removed.");
							ruleOwner.Rule = replacedSymbol;
						}

						if (ruleOwner.Type == SyntaxRuleType.Exception)
						{
							if (indexes.Contains(1))
							{
								var exceptionRule = (ExceptionSyntaxRule)ruleOwner;
								Debug.Assert(exceptionRule.Exception == null, "The rule should have been removed.");
								exceptionRule.Exception = replacedSymbol;
							}
						}

						continue;
					}

					var rulesOwner = rule as SyntaxRulesOwnerBase;
					if (rulesOwner != null)
					{
						var childRules = new List<SyntaxRule>(rulesOwner.Rules);
						rulesOwner.Rules.Clear();

						foreach (var index in indexes)
							childRules.Insert(index, replacedSymbol);

						foreach (var childRule in childRules)
							rulesOwner.Rules.Add(childRule);

						continue;
					}

					Debug.Fail("Unexpected");
				}
			}
		}
	}
}
